#ifndef _logger_H_
#define _logger_H_

#include <iostream>
#include <string>
#include <libscterm.hpp>
#include <mutex>

namespace sc {

    enum class loglevel {
        undefined,
        debug,
        info,
        warning,
        error,
        critical
    };

    constexpr bool operator>=(loglevel lhs, loglevel rhs) {
        return static_cast<int>(lhs) >= static_cast<int>(rhs);
    }

    constexpr const char* loglevel_desc(loglevel lvl) {
        switch (lvl) {
            case loglevel::undefined: return "UNDEFINED"; break;
            case loglevel::debug: return "DEBUG"; break;
            case loglevel::info: return "INFO"; break;
            case loglevel::warning: return "WARNING"; break;
            case loglevel::error: return "ERROR"; break;
            case loglevel::critical: return "CRITICAL"; break;
        }
        return "";
    }

    class logger {
        public:
            logger(const std::string& name, loglevel level);

            void msg(loglevel level, const sc::io::trgb* color,
                     const char* file, int line, const char* func);
           
            template<typename T, typename... Args>
            void msg(loglevel level, const sc::io::trgb* color,
                     const char* file, int line, const char* func,
                     T value, Args... args) {
                if (level >= _level) {
                    std::lock_guard<std::recursive_mutex> lock {_mtx};
                    if (_topcall) {
                        _topcall = false;
                        write_prefix(level, color, file, line, func);
                    }
                    std::cerr << value;
                    msg(level, color, file, line, func, args...);
                }
            }

            loglevel level() const { return _level; }
            void level(loglevel level) { _level = level; }

        private:
            void write_prefix(loglevel level,
                              const sc::io::trgb* color,
                              const char* file,
                              int line,
                              const char* func);
           
            loglevel _level {loglevel::info};
            bool _topcall {true};
            std::string _name;
            static std::recursive_mutex _mtx;
    };

}

#define SCDebug(logger, ...) \
    logger.msg(sc::loglevel::debug, nullptr, __FILE__, __LINE__, __func__, __VA_ARGS__)
#define SCInfo(logger, ...) \
    logger.msg(sc::loglevel::info, nullptr, __FILE__, __LINE__, __func__, __VA_ARGS__)
#define SCWarning(logger, ...) \
    logger.msg(sc::loglevel::warning, nullptr, __FILE__, __LINE__, __func__, __VA_ARGS__)
#define SCError(logger, ...) \
    logger.msg(sc::loglevel::error, nullptr, __FILE__, __LINE__, __func__, __VA_ARGS__)
#define SCCritical(logger, ...) \
    logger.msg(sc::loglevel::critical, nullptr, __FILE__, __LINE__, __func__, __VA_ARGS__)

#define SCDebugC(logger, color, ...) \
    logger.msg(sc::loglevel::debug, &color, __FILE__, __LINE__, __func__, __VA_ARGS__)
#define SCInfoC(logger, color, ...) \
    logger.msg(sc::loglevel::info, &color, __FILE__, __LINE__, __func__, __VA_ARGS__)
#define SCWarningC(logger, color, ...) \
    logger.msg(sc::loglevel::warning, &color, __FILE__, __LINE__, __func__, __VA_ARGS__)
#define SCErrorC(logger, color, ...) \
    logger.msg(sc::loglevel::error, &color, __FILE__, __LINE__, __func__, __VA_ARGS__)
#define SCCriticalC(logger, color, ...) \
    logger.msg(sc::loglevel::critical, &color, __FILE__, __LINE__, __func__, __VA_ARGS__)

#endif // _logger_H_
